-- MIT LICENSE
-- Copyright (c) 2016 Steve Wolligandt
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-- OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package.path = package.path .. ";../?.lua"
require "sampling"
require "rendering"
require "mathext"

windowName = "Newton Raphson 2D Vectorfield"
cursorVectorColor = { 58,137,201}
vectorFieldColor  = {242,108, 79}
maximumColor      = {255,0,0}
minimumColor      = {0,255,0}
saddleColor       = {210,210,0}
swirlingColor     = {0,0,0}
streamLineColor   = {200,200,200}

domain = {x={0,1,2},y={0,1,2}}
vf = {
    {{-0.25,0.6666666666},{0.75,0.6666666666},{0.75,-0.6666666666}},
    {{-0.25,-0.3333333333},{0.75,-0.1},{0.75,0.1}},
    {{-0.25,-0.3333333333},{-0.3,-0.1},{0.1,-0.6}}
}
function drawGraphs()
    -- draw vector field at all samples given
    love.graphics.setColor(vectorFieldColor)
    for x = 1,#domain.x do
        for y = 1,#domain.y do
            love.graphics.line(domain.x[x],domain.y[y],f(domain.x[x],domain.y[y])[1]+domain.x[x],-f(domain.x[x],domain.y[y])[2]+domain.y[y])
        end
    end

    -- coordinate of cursor in domainspace
    local domainPos = {
        x = (love.mouse.getX() - translation.x)/(scale),
        y = (love.mouse.getY() - translation.y)/(scale)
    }

    -- draw streamline by integrating in forward in backward stepsize 
    love.graphics.setColor(streamLineColor)
    local forward = {domainPos.x, domainPos.y}
    local backward = {domainPos.x, domainPos.y}

    local curForward = {domainPos.x, domainPos.y}
    local curBackward = {domainPos.x, domainPos.y}

    for i=1,500 do
        curForward = vector.rungeKutta4_2D(f,curForward, 0.1)
        curBackward = vector.rungeKutta4_2D(f,curBackward, -0.1)

        forward[#forward+1] = curForward[1]
        forward[#forward+1] = curForward[2]
        backward[#backward+1] = curBackward[1]
        backward[#backward+1] = curBackward[2]
    end
    love.graphics.line(forward)
    love.graphics.line(backward)

    --draw sampled vector at cursor position
    love.graphics.setLineWidth(love.graphics.getLineWidth()*20)
    love.graphics.setColor(cursorVectorColor)
    love.graphics.line(domainPos.x,domainPos.y,f(domainPos.x,domainPos.y)[1]+domainPos.x,-f(domainPos.x,domainPos.y)[2]+domainPos.y)

    -- draw extremal points
    love.graphics.setColor(0,0,255)
    for i = 1,#roots do
        love.graphics.setColor(roots[i].color)
        love.graphics.points(roots[i].coord[1],roots[i].coord[2])
    end
end

function love.mousepressed( x, y, button, istouch )
    local domainPos = {
        x = (x - translation.x)/(scale),
        y = (y - translation.y)/(scale)
    }

    local vector = f(domainPos.x, domainPos.y)
    print('f(' .. domainPos.x .. ', ' .. domainPos.y .. ') = (' .. vector[1] .. ', ' .. vector[2] .. ')')
end

function printtext()
    love.graphics.origin()
    love.graphics.setColor(cursorVectorColor)
    love.graphics.print('sampled vector', 10,10)
    love.graphics.setColor(vectorFieldColor)
    love.graphics.print('given samples', 10,30)
    love.graphics.setColor(maximumColor)
    love.graphics.print('source', 10,50)
    love.graphics.setColor(minimumColor)
    love.graphics.print('sink', 10,70)
    love.graphics.setColor(saddleColor)
    love.graphics.print('saddle', 10,90)
    love.graphics.setColor(swirlingColor)
    love.graphics.print('swirling', 10,110)
    love.graphics.setColor(streamLineColor)
    love.graphics.print('streamline', 10,130)
end

-- discrete function
function f(x,y)
    return vector.sampleDiscrete2D(x,y, domain, vf)
end 

-- derivative of discrete function
function df(x,y)
    return vector.discreteDerivative2D(x,y, domain, vf)
end

function createRandomVF() 
    domain = {x={-2,-1,0,1,2},y={-2,-1,0,1,2}}
    domain = {x = {},y={}}
    local i = 1
    for x = -10,10 do domain.x[i] = x; i = i+1 end
    local i = 1
    for y = -10,10 do domain.y[i] = y; i = i+1 end

    vf = {}
    for y=1,#domain.y do
        vf[y] = {}
        for x=1,#domain.x do
            vf[y][x] = {math.random()*0.8-0.4,math.random()*0.8-0.4}
        end
    end

    roots = {}
    for x = 1,#domain.x-1 do
        for y = 1,#domain.y-1 do
            for i = 0.0,1,0.2 do
                for j = 0.0,1,0.2 do
                    local root = vector.newton2D(f, df, {domain.x[x+1]*i + domain.x[x]*(1-i), domain.y[y+1]*j + domain.y[y]*(1-j)})
                    if (math.finite(root[1]) and math.finite(root[2])) then
                        local fRoot = f(root[1],root[2])
                        -- print('f(' .. root[1] .. ', ' .. root[2] .. ') = (' .. fRoot[1] .. ', ' .. fRoot[2]  .. ')')
                        if (domain.x[x] <= root[1] and domain.x[x+1] >= root[1] and domain.y[y] <= root[2] and domain.y[y+1] >= root[2] and math.abs(fRoot[1]) < 0.00000000001 and math.abs(fRoot[2]) < 0.0000001) then
                            roots[#roots+1] = {coord = root}
                            local eigVals = matrix.eig2x2(df(roots[#roots].coord[1], roots[#roots].coord[2])).values
                            if (eigVals[1] <= 0 and eigVals[2] <= 0) then
                                roots[#roots].color = maximumColor
                            elseif (eigVals[1] >= 0 and eigVals[2] >= 0 ) then
                                roots[#roots].color = minimumColor
                            elseif ((eigVals[1] >= 0 and eigVals[2] <= 0)) then
                                roots[#roots].color = saddleColor
                            else
                                roots[#roots].color = swirlingColor
                            end
                        end
                    end
                end
            end
        end
    end
end

function love.keypressed( key, scancode, isrepeat )
    if (key == 'space') then
        createRandomVF()
    end
end

function load()
    createRandomVF()
    love.window.setTitle(windowName)
    love.window.setMode( 1000, 1000, {} )
    love.graphics.setBackgroundColor(255,255,255)
    love.mouse.setVisible(false)
end
