-- MIT LICENSE
-- Copyright (c) 2016 Steve Wolligandt
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-- OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

scalar = {}
vector = {}
matrix = {}

-- simple linear interpolation
-- left is zero, right is 1
function scalar.interpolateLinear(left,right,i)
    return left*(1-i) + right*i
end

-- simple linear interpolation
-- left is zero, right is 1
function vector.interpolateLinear(left,right,i)
    return vector.add(vector.scale(left, (1-i)), vector.scale(right, (i)))
end

-- bilinear interpolation
-- bottomLeft is 0,0, topRight is 1,1
function scalar.interpolateBilinear(bottomLeft,bottomRight,topLeft,topRight, x,y) return interpolateBilinear(bottomLeft,bottomRight,topLeft,topRight, x,y, scalar) end
function vector.interpolateBilinear(bottomLeft,bottomRight,topLeft,topRight, x,y) return interpolateBilinear(bottomLeft,bottomRight,topLeft,topRight, x,y, vector) end
function interpolateBilinear(bottomLeft,bottomRight,topLeft,topRight, x,y, dimension)
    local bottom = dimension.interpolateLinear(bottomLeft, bottomRight, x)
    local top    = dimension.interpolateLinear(topLeft, topRight, x)

    return dimension.interpolateLinear(bottom, top, y)
end

-- bilinear interpolation
-- bottomLeftNear is 0,0,0, topRightFar is 1,1,1
function scalar.interpolateTrilinear(bottomLeftNear,bottomRightNear,topLeftNear,topRightNear, bottomLeftFar,bottomRightFar,topLeftFar,topRightFar, x,y,z) return interpolateTrilinear(bottomLeftNear,bottomRightNear,topLeftNear,topRightNear, bottomLeftFar,bottomRightFar,topLeftFar,topRightFar, x,y,z, scalar) end
function vector.interpolateTrilinear(bottomLeftNear,bottomRightNear,topLeftNear,topRightNear, bottomLeftFar,bottomRightFar,topLeftFar,topRightFar, x,y,z) return interpolateTrilinear(bottomLeftNear,bottomRightNear,topLeftNear,topRightNear, bottomLeftFar,bottomRightFar,topLeftFar,topRightFar, x,y,z, vector) end
function interpolateTrilinear(bottomLeftNear,bottomRightNear,topLeftNear,topRightNear, bottomLeftFar,bottomRightFar,topLeftFar,topRightFar, x,y,z, dimension)
    local bottomLeft  = dimension.interpolateLinear(bottomLeftNear,  bottomLeftFar, z)
    local bottomRight = dimension.interpolateLinear(bottomRightNear, bottomRightFar, z)
    local topLeft     = dimension.interpolateLinear(topLeftNear,     topLeftFar,     z)
    local topRight    = dimension.interpolateLinear(topRightNear,    topRightFar,    z)

    return dimension.interpolateBilinear(bottomLeft, bottomRight, topLeft, topRight, x,y)
end

-- derivate of discrete function by local neighbour changes
function scalar.discreteDerivative1D(x, domain, functionvalues)
    return scalar.sampleDiscrete1D(x+1, domain, functionvalues) - scalar.sampleDiscrete1D(x-1, domain, functionvalues)
end

-- derivate of discrete function by local neighbour changes
function scalar.discreteDerivative2D(x,y, domain, functionvalues)
    return {
        scalar.sampleDiscrete2D(x+1,y, domain, functionvalues) - scalar.sampleDiscrete2D(x-1,y, domain, functionvalues),
        scalar.sampleDiscrete2D(x,y+1, domain, functionvalues) - scalar.sampleDiscrete2D(x,y-1, domain, functionvalues)
    }
end

-- derivate of discrete function by local neighbour changes
function scalar.discreteDerivative3D(x,y,z, domain, functionvalues)
    return {
        scalar.sampleDiscrete2D(x+1,y,z, domain, functionvalues) - scalar.sampleDiscrete2D(x-1,y,z, domain, functionvalues),
        scalar.sampleDiscrete2D(x,y+1,z, domain, functionvalues) - scalar.sampleDiscrete2D(x,y-1,z, domain, functionvalues),
        scalar.sampleDiscrete2D(x,y,z+1, domain, functionvalues) - scalar.sampleDiscrete2D(x,y,z-1, domain, functionvalues)
    }
end

-- derivate of discrete function by local neighbour changes
function vector.discreteDerivative1D(x, domain, functionvalues)
    local left  = vector.sampleDiscrete1D(x-1, domain, functionvalues)
    local right = vector.sampleDiscrete1D(x+1, domain, functionvalues)
    return vector.scale(vector.subtract(right,left),0.5)
end

-- derivate of discrete function by local neighbour changes
function vector.discreteDerivative2D(x,y, domain, functionvalues)
    local left   = vector.sampleDiscrete2D(x-0.00001,y, domain, functionvalues)
    local right  = vector.sampleDiscrete2D(x+0.00001,y, domain, functionvalues)
    local bottom = vector.sampleDiscrete2D(x,y-0.00001, domain, functionvalues)
    local top    = vector.sampleDiscrete2D(x,y+0.00001, domain, functionvalues)

    return {
        vector.scale(vector.subtract(right,left),50000), 
        vector.scale(vector.subtract(top,bottom),50000)
    }
end

-- derivate of discrete function by local neighbour changes
function vector.discreteDerivative3D(x,y,z, domain, functionvalues)
    local left   = vector.sampleDiscrete3D(x-1,y,z, domain, functionvalues)
    local right  = vector.sampleDiscrete3D(x+1,y,z, domain, functionvalues)
    local bottom = vector.sampleDiscrete3D(x,y-1,z, domain, functionvalues)
    local top    = vector.sampleDiscrete3D(x,y+1,z, domain, functionvalues)
    local near   = vector.sampleDiscrete3D(x,y,z-1, domain, functionvalues)
    local far    = vector.sampleDiscrete3D(x,y,z+1, domain, functionvalues)

    return {
        vector.scale(vector.subtract(right,left),0.5),
        vector.scale(vector.subtract(top,bottom),0.5),
        vector.scale(vector.subtract(far,near),0.5)
    }
end

-- samples a discrete function at x
function scalar.sampleDiscrete1D(x, domain, functionvalues) return sampleDiscrete1D(x, domain, functionvalues, scalar) end
function vector.sampleDiscrete1D(x, domain, functionvalues) return sampleDiscrete1D(x, domain, functionvalues, vector) end
function sampleDiscrete1D(x, domain, functionvalues, dimension)
    -- clamp to border
    if (x <= domain[1]) then 
        return functionvalues[1] 
    end
    if (x >= domain[#domain] ) then 
        return functionvalues[#functionvalues] 
    end

    for i=1,(#domain-1) do
        -- on sample
        if (x == domain[i]) then return functionvalues[i] end

        -- between samples
        if (x > domain[i] and x < domain[i+1]) then
            return dimension.interpolateLinear(functionvalues[i], functionvalues[i+1], (x-domain[i])/(domain[i+1]-domain[i]))
        end
    end
end

-- samples a discrete function at x,y
function scalar.sampleDiscrete2D(x,y, domain, functionvalues) return sampleDiscrete2D(x,y,domain,functionvalues,scalar) end
function vector.sampleDiscrete2D(x,y, domain, functionvalues) return sampleDiscrete2D(x,y,domain,functionvalues,vector) end
function sampleDiscrete2D(x,y, domain, functionvalues, dimension)
    -- clamp coordinate
    local clampedX = x
    local clampedY = y

    if (x < domain.x[1]) then clampedX = domain.x[1] end
    if (x > domain.x[#domain.x] ) then clampedX = domain.x[#domain.x] end

    if (y < domain.y[1]) then clampedY = domain.y[1] end
    if (y > domain.y[#domain.y] ) then clampedY = domain.y[#domain.y] end

    -- direct sample positions
    for curX=1,(#domain.x) do
        for curY=1,(#domain.y) do
            if (clampedX == domain.x[curX] and clampedY == domain.y[curY]) then return functionvalues[curY][curX] end
        end
    end

    -- check each quad, if x,y is in quad, bilinear interpolate
    for curX=1,(#domain.x-1) do
        for curY=1,(#domain.y-1) do
            if (clampedX >= domain.x[curX] and clampedX <= domain.x[curX+1] and clampedY >= domain.y[curY] and clampedY <= domain.x[curY+1]) then
                return dimension.interpolateBilinear(
                    functionvalues[curY][curX], functionvalues[curY][curX+1], 
                    functionvalues[curY+1][curX], functionvalues[curY+1][curX+1], 
                    (clampedX-domain.x[curX])/(domain.x[curX+1]-domain.x[curX]), 
                    (clampedY-domain.y[curY])/(domain.y[curY+1]-domain.y[curY]))
            end
        end
    end
end

-- samples a discrete function at x,y
function scalar.sampleDiscrete3D(x,y,z, domain, functionvalues) return sampleDiscrete3D(x,y,z,domain,functionvalues, scalar) end
function vector.sampleDiscrete3D(x,y,z, domain, functionvalues) return sampleDiscrete3D(x,y,z,domain,functionvalues, vector) end
function sampleDiscrete3D(x,y,z, domain, functionvalues, dimension)
    -- clamp coordinate
    local clampedX = x
    local clampedY = y
    local clampedZ = z
    
    if (x < domain.x[1]) then clampedX = domain.x[1] end
    if (x > domain.x[#domain.x] ) then  clampedX = domain.x[#domain.x] end

    if (y < domain.y[1]) then clampedY = domain.y[1] end
    if (y > domain.y[#domain.y] ) then  clampedY = domain.y[#domain.y] end

    if (z < domain.z[1]) then clampedZ = domain.z[1] end
    if (z > domain.z[#domain.z] ) then clampedZ = domain.z[#domain.z] end

    -- direct sample positions
    for curX=1,(#domain.x) do
        for curY=1,(#domain.y) do
            for curZ=1,(#domain.z) do
                if (clampedX == domain.x[curX] and clampedY == domain.y[curY] and clampedZ == domain.z[curZ]) then return functionvalues[curZ][curY][curX] end
            end
        end
    end

    -- check each cell, if x,y,z is in cell, trilinear interpolate
    for curX=1,(#domain.x-1) do
        for curY=1,(#domain.y-1) do
            for curZ=1,(#domain.z-1) do
                if (clampedX >= domain.x[curX] and clampedX <= domain.x[curX+1] and clampedY >= domain.y[curY] and clampedY <= domain.y[curY+1] and clampedZ >= domain.z[curZ] and clampedZ <= domain.z[curZ+1]) then
                    return dimension.interpolateTrilinear(
                        functionvalues[curZ]  [curY][curX], functionvalues[curZ]  [curY][curX+1], functionvalues[curZ]  [curY+1][curX], functionvalues[curZ]  [curY+1][curX+1], 
                        functionvalues[curZ+1][curY][curX], functionvalues[curZ+1][curY][curX+1], functionvalues[curZ+1][curY+1][curX], functionvalues[curZ+1][curY+1][curX+1], 
                        (clampedX-domain.x[curX])/(domain.x[curX+1]-domain.x[curX]), 
                        (clampedY-domain.y[curY])/(domain.y[curY+1]-domain.y[curY]), 
                        (clampedZ-domain.z[curZ])/(domain.z[curZ+1]-domain.z[curZ]))
                end
            end
        end
    end
end

function vector.rungeKutta4_2D(f, coord, stepSize)


    local k1 = f(coord[1],coord[2])

    local k2Coord = vector.add(coord, vector.scale(k1,stepSize * 0.5))
    local k2 = f(k2Coord[1], k2Coord[2])

    local k3Coord = vector.add(coord, vector.scale(k2,stepSize * 0.5))
    local k3 = f(k3Coord[1], k3Coord[2]);

    local k4Coord = vector.add(coord, vector.scale(k3,stepSize * 0.5))
    local k4 = f(k4Coord[1], k4Coord[2]);
    return vector.add(coord, vector.scale(vector.add(vector.add(vector.scale(k1,1/6),  vector.scale(k2,1/3)), vector.add(vector.scale(k3,1/3), vector.scale(k4,1/6))), stepSize));
end


-- newton raphson method for finding roots of nonlinear functions
function scalar.newton1D(func, derivative, coord)
    local old = coord
    local notequal = true
    while notequal do
        local new = old - func(old) / derivative(old)
        notequal = math.abs(old-new) > 0.000000000000000000000001
        old = new
    end

    return old
end

-- newton raphson method for finding roots of nonlinear functions
function vector.newton2D(func, derivative, coord)
    local old = {coord[1], coord[2]}
    local new = {coord[1], coord[2]}
    local notequal = true
    local numIt = 0
    while (notequal and numIt < 10) do
        new = vector.subtract(old, matrix.vectorMultiply(matrix.inverse2x2(derivative(old[1], old[2])), func(old[1], old[2])))
        notequal = (math.abs(old[1]-new[1]) + math.abs(old[2]-new[2])) > 0.0000000000000000001
        old = {new[1], new[2]}
        numIt=numIt+1
    end

    return new
end

-- printing a vector
function vector.print(vec)
    local str = "("
    for i=1,#vec do 
        str = str .. vec[i] 
        if (i~=#vec) then str = str .. "|" end
    end
    str = str .. ")"

    print(str)
end

-- printing a vector
function matrix.print(mat)
    for x=1,#mat[1] do 
        local str = "("
        for y=1,#mat do 
            str = str .. mat[y][x] 
            if (y~=#mat) then str = str .. "  " end
        end
        str = str .. ")"
        print(str)
    end

end

function vector.add(vec1, vec2)
    assert(#vec1 == #vec2, "## function vector.add(vec1, vec2) ##\ndimensions differ: length(vec1) = " .. #vec1 .. ", length(vec2) = " .. #vec2)
    local added = {}
    for i=1,#vec1 do 
        added[i] = vec1[i]+vec2[i]
    end
    return added
end

function vector.negate(vec)
    local negated = {}
    for i=1,#vec do 
        negated[i] = -vec[i]
    end
    return negated
end

function vector.subtract(vec1, vec2)
    assert(#vec1 == #vec2, "## function vector.subtract(vec1, vec2) ##\ndimensions differ: length(vec1) = " .. #vec1 .. ", length(vec2) = " .. #vec2)
    return vector.add(vec1, vector.negate(vec2))
end

function vector.scale(vec, factor)
    local scaled = {}
    for i=1,#vec do 
        scaled[i] = vec[i]*factor
    end
    return scaled
end

function vector.multiplyComponentwise(vec1, vec2)
    assert(#vec1 == #vec2, "## function vector.multiplyComponentwise(vec1, vec2) ##\ndimensions differ: length(vec1) = " .. #vec1 .. ", length(vec2) = " .. #vec2)
    local mult = {}
    for i=1,#vec1 do 
        mult[i] = vec1[i]*vec2[i]
    end
    return mult
end

function vector.invert(vec)
    local inverted = {}
    for i=1,#vec do 
        inverted[i] = 1/vec[i]
    end
    return inverted
end

function vector.divideComponentwise(vec1, vec2)
    assert(#vec1 == #vec2, "## function vector.divideComponentwise(vec1, vec2) ##\ndimensions differ: length(vec1) = " .. #vec1 .. ", length(vec2) = " .. #vec2)
    return vector.multiplyComponentwise(vec1, vector.invert(vec2))
end

-- returns table with field values and field vectors
-- matrix is stored columnwise
function matrix.eig2x2(mat)
    assert(#mat == 2 and #mat[1] == 2, "## unction matrix.eig2x2(mat) ##\nmatrix dimensions do not match: "..#mat.."x"..#mat[1])
    local eigenanalysis = {}
    eigenanalysis.vectors = {}

    local a = mat[1][1];
    local b = mat[2][1];
    local c = mat[1][2];
    local d = mat[2][2];

    local pHalf = (a+d)*0.5;
    local pHalfSqr = pHalf*pHalf;
    local q = a*d-b*c;

    local lambda1 = pHalf + math.sqrt(pHalfSqr - q)
    local lambda2 = pHalf - math.sqrt(pHalfSqr - q)

    eigenanalysis.values = {math.max(lambda1, lambda2), math.min(lambda1, lambda2)}
    return eigenanalysis
end

function matrix.transpose(mat)
    local trans = {}
    for i=1,#mat[1] do
        trans[i] = {}
    end

    for x=1,#mat do
        for y=1,#mat[1] do
            trans[y][x] = mat[x][y]
        end
    end

    return trans
end

function matrix.matrixMultiply(matLeft, matRight)
    assert(#matLeft == #matRight[1], "## function matrix.matrixMultiply(matLeft, matRight) ##\n")

    local mult = {}
    for yLeft = 1,#matLeft[1] do
        mult[yLeft] = {}
        for xRight = 1,#matRight do
            mult[yLeft][xRight] = 0
        end
    end

    for yLeft=1,#matLeft[1] do
        for xRight=1,#matRight do
            for i=1,#matLeft do
                mult[xRight][yLeft] = mult[xRight][yLeft] + matLeft[i][yLeft]*matRight[xRight][i]
            end
        end
    end

    return mult
end

function matrix.vectorMultiply(matLeft, vecRight)
    assert(#matLeft == #vecRight, "## functionmatrix.vectorMultiply(matLeft, vecRight) ##\n")

    local mult = {}
    for i = 1,#matLeft[1] do
        mult[i] = 0
    end

    for yLeft=1,#matLeft[1] do
        for i=1,#matLeft do
            mult[yLeft] = mult[yLeft] + matLeft[i][yLeft]*vecRight[i]
        end
    end

    return mult
end

function matrix.inverse2x2(mat)
    local factor = 1/(mat[1][1]*mat[2][2]-mat[2][1]*mat[1][2])
    local inverse = {{mat[2][2]*factor,-mat[1][2]*factor},{-mat[2][1]*factor,mat[1][1]*factor}}
    return inverse
end
